import { curry } from './curry.js'
import { inc } from './math.js'

/**
 * @template T
 * @param {T | Iterable<T> | Iterator<T>} value
 * @returns {value is Iterable<T>}
 */
export function isIterable(value) {
  return value[Symbol.iterator] != null
}

/**
 * @template T
 * @param {T | Iterable<T> | Iterator<T>} value 
 * @yields {T}
 */
export function* iter(value) {
  if (isIterable(value)) {
    yield* Iterator.from(value)
  } else {
    yield value
  }
}

/**
 * @template T
 * @param {...(Iterable<T> | Iterator<T>)} iterators
 * @yields {T}
 */
export const concat = function*(...iterators) {
  for (const iter of iterators) {
    for (const item of Iterator.from(iter)) {
      yield item
    }
  }
}
/**
 * @template T
 * @param {number} i
 * @param {T[] | Iterable<T>} a
 * @returns {T}
 */
export const nth = (i, a) => {
  if (Array.isArray(a)) {
    return a[i]
  } else {
    return /** @type {T} */ (iter(a).find((_, n) => i === n))
  }
}


/**
 * @template T
 * @param {number} n
 * @param {T[] | Iterable<T>} a
 * @returns {a}
 */
export const take = (n, a) => {
  if ('slice' in a) {
    return a.slice(0, n)
  } else {
    iter(a).take(n)
    return a
  }
}

/**
 * @template T
 * @param {number} n
 * @param {T[] | Iterable<T>} a
 * @returns {a}
 */
export const drop = (n, a) => {
  if ('slice' in a) {
    return a.slice(n)
  } else {
    iter(a).drop(n)
    return a
  }
}

/**
 * @param {ArrayLike} a
 * @returns {number}
 */
export const length = a => {
  if ('length' in a) {
    return a.length
  } else {
    return iter(a).reduce(inc, 0)
  }
}

/**
 * @template T
 * @param {T[] | Iterable<T>} a
 * @returns {T}
 */
export const head = a => nth(0, a)

/**
 * @template T
 * @param {T[] | Iterable<T>} a
 * @returns {a}
 */
export const tail = a => drop(1, a)

/**
 * @template T
 * @param {T[]} a
 */
export const last = a => nth(length(a) - 1, a)

export const prepend = curry(
  /**
    * @template T
    * @param {T} x 
    * @param {Iterable<T>} xs 
    */
  (x, xs) => concat([x], xs))


